from PySide6 import QtWidgets, QtCore, QtGui
from meni.ui.common import DockTitleBar, IconLabel
from meni.ui.propertyrowwithapply import CollectionPropertyRowWithApply
from meni.utils import tags_from_text

import qtawesome as qta


class CollectionPropertiesDock(QtWidgets.QDockWidget):
    def __init__(self, parent):
        super().__init__("Collection Properties", objectName="properties", parent=parent)

        self.app = QtCore.QCoreApplication.instance()

        self.setTitleBarWidget(DockTitleBar("Collection Properties", clicked=self.close))

        self.layout = QtWidgets.QFormLayout()

        self.empty = QtWidgets.QLabel("No collection selected")
        self.empty.setAlignment(QtCore.Qt.AlignCenter)
        self.empty.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        self.empty.setStyleSheet("color: rgba(255,255,255,0.2);")
        self.layout.addRow(self.empty)

        # === Name
        self.col_name = QtWidgets.QLineEdit()
        self.col_name.returnPressed.connect(lambda: self.apply_collection(name=self.col_name.text()))
        row = self.layout.addRow(
            "Name",
            CollectionPropertyRowWithApply(
                self,
                self.col_name,
                lambda: dict(name=self.col_name.text()),
                has_apply_all=False,
            ),
        )

        # === Author
        self.col_author = QtWidgets.QLineEdit()
        self.col_author.returnPressed.connect(lambda: self.apply_collection(author=self.col_author.text()))
        self.layout.addRow(
            "Author",
            CollectionPropertyRowWithApply(
                self,
                self.col_author,
                lambda: dict(author=self.col_author.text()),
            ),
        )

        # === URL
        self.col_url = QtWidgets.QLineEdit()
        self.col_url.returnPressed.connect(lambda: self.apply_collection(url=self.col_url.text()))
        self.layout.addRow(
            "URL",
            CollectionPropertyRowWithApply(
                self,
                self.col_url,
                lambda: dict(url=self.col_url.text()),
            ),
        )

        # === Notes
        self.col_notes = QtWidgets.QPlainTextEdit()
        self.layout.addRow(
            "Notes",
            CollectionPropertyRowWithApply(
                self,
                self.col_notes,
                lambda: dict(notes=self.col_notes.toPlainText()),
            ),
        )

        # Set layout to widget to dock
        widget = QtWidgets.QWidget()
        widget.setLayout(self.layout)
        self.setWidget(widget)

        # Connect signals
        self.app.metadata.changed.connect(lambda: self.refresh([], self.app.last_selected_file))
        self.app.selected_files_changed.connect(self.refresh)

        self.refresh([], None)

    def refresh(self, files, last):
        collection = last and last.collection_obj
        if collection is not None:
            self.col_name.setText(collection.name)
            self.col_author.setText(collection.author)
            self.col_url.setText(collection.url)
            self.col_notes.setPlainText(collection.notes)

        self.layout.setRowVisible(0, collection is None)
        for i in range(1, self.layout.rowCount()):
            self.layout.setRowVisible(i, collection is not None)

    def apply_collection(self, **kwargs):
        if self.app.last_selected_file and self.app.last_selected_file.collection_obj:
            self.app.metadata.update_collection(self.app.last_selected_file.collection_obj, **kwargs)



from PySide6 import QtWidgets, QtCore
from meni.ui.windows.mainwindow import MainWindow
from meni.ui.windows.welcome import WelcomeWindow
from meni.model.model import JsonPickledMetadata
from meni.theme import *


class AppMeni(QtWidgets.QApplication):

    status = QtCore.Signal(str)
    filter_changed = QtCore.Signal()
    selected_files_changed = QtCore.Signal(list, object)

    def __init__(self, sys_argv, library=None):
        super().__init__(sys_argv)

        self.setApplicationName("Meni 3D Library")

        self._library_command_line = library

        self.settings = QtCore.QSettings("meni", "meni")
        # self.threadpool = QtCore.QThreadPool()
        # self.threadpool.setMaxThreadCount(1)
        self.metadata = JsonPickledMetadata()

        self.main = None
        self.welcome = None
        self._search_filter = None
        self._selected_files = []
        self.filters = []

        self.theme = Nord()
        # self.theme = Dracula()
        # self.theme = Gruvbox()

        self.status.connect(print)

    def startup(self):
        if self.current_library:
            self.show_main()
        else:
            self.show_welcome()

    def show_main(self):
        if not self.main:
            self.main = MainWindow()
        self.main.show()

    def show_welcome(self):
        if not self.welcome:
            self.welcome = WelcomeWindow()
        self.welcome.show()

    def add_filter(self, filter):
        self.filters.append(filter)

    @property
    def current_library(self):
        return self._library_command_line or self.settings.value("current_library")

    @current_library.setter
    def current_library(self, value):
        self.settings.setValue("current_library", value)
        self._library_command_line = None
        self.metadata.reload()

    @property
    def last_selected_file(self):
        if len(self._selected_files) > 0:
            return self._selected_files[-1]
        return None

    @property
    def selected_files(self):
        return self._selected_files

    @selected_files.setter
    def selected_files(self, list):
        self._selected_files = list
        self.selected_files_changed.emit(list, list[-1] if list else None)

    @property
    def search_filter(self):
        return self._search_filter

    @search_filter.setter
    def search_filter(self, value):
        self._search_filter = value
        self.filter_changed.emit()



from PySide6 import QtWidgets, QtCore, QtGui
import qtawesome as qta
import subprocess
import os


class FileContextMenu(QtWidgets.QMenu):
    def __init__(self, parent, files):
        super().__init__(parent)
        self.files = files
        self.app = QtCore.QCoreApplication.instance()

        remove_action = QtGui.QAction(
            f"Remove from library", self, icon=qta.icon("fa5s.minus-square", color=self.app.theme.icon_color), text="Remove from library"
        )
        remove_action.triggered.connect(self.remove_file)
        self.addAction(remove_action)

        open_action = QtGui.QAction(
            f"Open with system default application",
            self,
            icon=qta.icon("fa5s.external-link-alt", color=self.app.theme.icon_color),
            text="Open with default application",
        )
        open_action.triggered.connect(self.open_with_system)
        self.addAction(open_action)

    def remove_file(self):
        dialog = QtWidgets.QMessageBox(self)
        dialog.setWindowTitle("Remove file")
        dialog.setIconPixmap(qta.icon("fa5.question-circle", color=self.app.theme.icon_color).pixmap(64, 64))

        deletion = self.files[0] if len(self.files) == 1 else f"{len(self.files)} files"
        dialog.setText(f"Are you sure you want to remove <b>{deletion}</b> from the library?<br><br>This will also remove the file(s) from disk.")
        dialog.setStandardButtons(QtWidgets.QMessageBox.StandardButton.Yes | QtWidgets.QMessageBox.StandardButton.No)
        dialog.setDefaultButton(QtWidgets.QMessageBox.StandardButton.No)

        if dialog.exec() == QtWidgets.QMessageBox.StandardButton.Yes:
            for file in self.files:
                self.app.metadata.remove_file(file)
                self.app.status.emit(f"Removed {file.name} from the library")

    def open_with_system(self):
        filename = self.files[0].path
        try:
            os.startfile(filename)
        except AttributeError:
            subprocess.call(["open", filename])



import os

from PySide6 import QtWidgets, QtCore, QtGui
from meni.ui.filestable import FilesTable
from meni.ui.searchinput import SearchInput
from meni.ui.windows.importdialog import ImportDialog
from meni.ui.menus.menusettings import MenuSettings
from meni.ui.docks.collectionproperties import CollectionPropertiesDock
from meni.ui.docks.fileproperties import FilePropertiesDock
from meni.ui.docks.viewer import ViewerDock
from meni.ui.docks.browser import BrowserDock
from meni.ui.toolbars.maintoolbar import MainToolbar
import qtawesome as qta


class MainWindow(QtWidgets.QMainWindow):
    def __init__(self):
        super().__init__()

        self.app = QtCore.QCoreApplication.instance()

        self.setStyleSheet(
            f"""
                * {{
                    background-color: {self.app.theme.main_background};
                    color: {self.app.theme.main_foreground};
                    selection-background-color: {self.app.theme.selection_background};
                    selection-color: {self.app.theme.selection_foreground};
                }}

                QMainWindow::separator {{
                    background-color: rgba(0, 0, 0, 0.15);
                    width: 4px;
                    border: none;
                }}

                QPushButton#save:enabled, QPushButton#Ok:enabled {{
                    background-color: {self.app.theme.green_btn};
                    color: #000;
                }}

                QPushButton#cancel:enabled {{
                    background-color: {self.app.theme.red_btn};
                    color: #000;
                }}
                
                QLineEdit, QComboBox, QTextEdit, QPlainTextEdit {{
                    background-color: rgba(255, 255, 255, 0.05);
                    border: 1px solid rgba(0, 0, 0, 0.3);
                }}

                QTableView QLineEdit, QTableView QComboBox {{
                    background-color: {self.app.theme.main_background};
                }}

                QTableView {{
                    gridline-color: rgba(0, 0, 0, 0.2);
                }}
            """
        )

        self.setWindowTitle("Meni 3D Library")
        self.setWindowFlags(QtCore.Qt.WindowType.Dialog)
        self.resize(self.app.settings.value("size", QtCore.QSize(270, 225)))
        self.move(self.app.settings.value("pos", QtCore.QPoint(50, 50)))

        self.toolbar = MainToolbar(self)
        self.addToolBar(self.toolbar)

        self.table = FilesTable()

        self.setCentralWidget(self.table)
        self.viewer = ViewerDock(self)
        self.fileproperties = FilePropertiesDock(self)
        self.collectionproperties = CollectionPropertiesDock(self)
        self.browser = BrowserDock(self)

        self.addDockWidget(QtCore.Qt.DockWidgetArea.RightDockWidgetArea, self.viewer)
        self.addDockWidget(QtCore.Qt.DockWidgetArea.RightDockWidgetArea, self.fileproperties)
        self.addDockWidget(QtCore.Qt.DockWidgetArea.RightDockWidgetArea, self.collectionproperties)
        self.addDockWidget(QtCore.Qt.DockWidgetArea.LeftDockWidgetArea, self.browser)

        self.app.status.connect(self.on_status)

        self.restoreState(self.app.settings.value("state"))

    def closeEvent(self, event):
        self.app.settings.setValue("size", self.size())
        self.app.settings.setValue("pos", self.pos())
        self.app.settings.setValue("state", self.saveState())
        event.accept()

    def on_status(self, message):
        self.statusBar().showMessage(message, 5000)



from PySide6 import QtWidgets, QtCore, QtGui
from meni.model.model import Local3DFile, Stage
from meni.ui.common import QHLine, QVLine
from meni.utils import tags_from_text
import os


class ImportDialog(QtWidgets.QDialog):
    def __init__(self, parent, files=None):
        super().__init__(parent)

        self.stage = Stage()
        self.app = QtCore.QCoreApplication.instance()

        self.setWindowTitle("Meni 3D Library - Import")

        hlayout = QtWidgets.QHBoxLayout()

        # Left panel
        self.left_panel = Panel("1. Add files", "Supported file format: STL. You can drag multiple files or a folder containing 3D models.")
        self.dndtarget = DragAndDropTarget()
        self.dndtarget.files_droped.connect(self.add_files)
        self.left_panel.add_widget(self.dndtarget)
        self.left_panel.add_widget(QtWidgets.QLabel("OR", alignment=QtCore.Qt.AlignmentFlag.AlignCenter))
        self.btn_browse_files = QtWidgets.QPushButton("Browse files...")
        self.btn_browse_files.clicked.connect(self.browse_files)
        self.left_panel.add_widget(self.btn_browse_files)

        hlayout.addWidget(self.left_panel)
        hlayout.addWidget(QVLine())

        # Middle panel
        self.middle_panel = Panel(
            "2. Rename the files",
            "Double-click to rename files as needed. You can always change the name later.",
        )
        self.file_list = AddedFilesTable()
        self.file_list.itemChanged.connect(self.on_edit)
        self.middle_panel.add_widget(self.file_list)

        hlayout.addWidget(self.middle_panel)
        hlayout.addWidget(QVLine())

        # Right panel
        self.right_panel = Panel(
            "3. Define collection and tags",
            "Define a collection and tags to apply to all imported files. You can modify them later.",
        )
        self.right_panel.add_widget(QtWidgets.QLabel("Collection name"))

        self.collection = QtWidgets.QComboBox(self)
        self.collection.addItems([col.name for col in self.app.metadata.collections])
        self.collection_edit = QtWidgets.QLineEdit(self)
        self.collection.setLineEdit(self.collection_edit)
        self.collection_edit.setText("")

        self.right_panel.add_widget(self.collection)
        self.right_panel.add_widget(QtWidgets.QLabel("Tags"))
        self.tags = QtWidgets.QLineEdit(self)
        self.right_panel.add_widget(self.tags)
        self.right_panel.layout.addStretch(1)

        hlayout.addWidget(self.right_panel)

        btns = QtWidgets.QDialogButtonBox.Ok | QtWidgets.QDialogButtonBox.Cancel
        self.buttonBox = QtWidgets.QDialogButtonBox(btns)
        self.buttonBox.accepted.connect(self.on_OK)
        self.buttonBox.rejected.connect(self.reject)
        self.buttonBox.button(QtWidgets.QDialogButtonBox.Ok).setEnabled(False)

        self.layout = QtWidgets.QVBoxLayout()
        self.layout.addLayout(hlayout)
        self.layout.addWidget(QHLine())
        self.layout.addWidget(self.buttonBox)
        self.setLayout(self.layout)

        self.add_files(files)

    def on_OK(self):
        self.stage.collection = self.collection.currentText()
        self.stage.tags = tags_from_text(self.tags.text())
        self.app.metadata.commit_stage(self.stage)
        self.accept()

    def add_files(self, files):
        if files:
            for file in files:
                self.add_file(file)

    def add_file(self, path):
        if os.path.isdir(path):
            for child in os.listdir(path):
                self.add_file(os.path.join(path, child))
        else:
            if path.lower().endswith(".stl"):
                file = Local3DFile(path, collection=None)
                self.stage.files.append(file)
                self.file_list.add_file(file)
                self.buttonBox.button(QtWidgets.QDialogButtonBox.Ok).setEnabled(True)

    def on_edit(self, item):
        file = self.stage.files[item.row()]
        if item.column() == 0:
            file.name = item.text()

    def browse_files(self):
        file_dialog = QtWidgets.QFileDialog()
        file_dialog.setWindowTitle("Meni 3D Library")
        file_dialog.setDirectory(self.app.settings.value("last_path", QtCore.QDir.homePath()))
        file_dialog.setFileMode(QtWidgets.QFileDialog.ExistingFiles)
        file_dialog.setNameFilter("3D Files (*.stl)")
        file_dialog.setViewMode(QtWidgets.QFileDialog.Detail)
        if file_dialog.exec_():
            paths = file_dialog.selectedFiles()
            for path in paths:
                self.app.settings.setValue("last_path", os.path.dirname(path))
                self.add_file(path)


class TitleLabel(QtWidgets.QLabel):
    def __init__(self, text):
        super().__init__(text)
        self.setStyleSheet("font-size: 15px; font-weight: bold; text-decoration: underline;")


class Panel(QtWidgets.QWidget):
    def __init__(self, title, description):
        super().__init__()

        self.setMinimumSize(300, 300)

        self.layout = QtWidgets.QVBoxLayout()
        self.setLayout(self.layout)

        self.layout.addWidget(TitleLabel(title))
        self.description = QtWidgets.QLabel(description, alignment=QtCore.Qt.AlignmentFlag.AlignTop)
        self.description.setWordWrap(True)
        self.description.setMinimumHeight(60)
        self.layout.addWidget(self.description)

    def add_widget(self, widget):
        self.layout.addWidget(widget)


class AddedFilesTable(QtWidgets.QTableWidget):
    def __init__(self):
        super().__init__()

        self.setColumnCount(2)
        self.setHorizontalHeaderLabels(["Name", ""])
        self.horizontalHeader().setSectionResizeMode(QtWidgets.QHeaderView.ResizeToContents)
        self.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)

    def add_file(self, file):
        row = self.rowCount()

        self.insertRow(row)
        self.setItem(row, 0, QtWidgets.QTableWidgetItem(file.name))
        self.setItem(row, 1, QtWidgets.QTableWidgetItem(file.path))

        # Make name editable
        item = self.item(row, 0)
        item.setFlags(item.flags() | QtCore.Qt.ItemFlag.ItemIsEditable)

        # Disable path edition
        item = self.item(row, 1)
        item.setFlags(item.flags() & ~QtCore.Qt.ItemFlag.ItemIsEditable)


class DragAndDropTarget(QtWidgets.QWidget):

    files_droped = QtCore.Signal(list)

    def __init__(self):
        super().__init__()

        self.setAcceptDrops(True)

        self.setStyleSheet(
            """
                           background-color: rgba(255, 255, 255, 0.1); 
                           padding: 30px; 
                           border: 2px dashed rgba(255, 255, 255, 0.2); 
                           border-radius: 5px;
                           font-size: 17px;
                           text-align: center;"""
        )
        layout = QtWidgets.QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(QtWidgets.QLabel("Drag your model files here", alignment=QtCore.Qt.AlignmentFlag.AlignCenter))

        self.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        self.setLayout(layout)

    def dragEnterEvent(self, event):
        if event.mimeData().hasUrls():
            event.accept()
        else:
            event.ignore()

    def dropEvent(self, event):
        files = []
        for url in event.mimeData().urls():
            files.append(url.toLocalFile())
        event.accept()
        self.files_droped.emit(files)



from PySide6 import QtWidgets, QtCore
from qt_material import apply_stylesheet
from meni.ui.windows.mainwindow import MainWindow
from meni.ui.windows.welcome import WelcomeWindow
from meni.app import AppMeni
import sys
import argparse


if __name__ == "__main__":

    parser = argparse.ArgumentParser(description="Meni: Library manager for 3D models and assets.")
    parser.add_argument("-l", "--library", help="Path to the library directory.")
    args = parser.parse_args()

    app = AppMeni(sys.argv, library=args.library)

    # app.setStyleSheet(qdarkstyle.load_stylesheet_pyside6())

    # setup stylesheet
    # apply_stylesheet(app, theme="dark_teal.xml")
    # extra = {"density_scale": "-3"}
    # apply_stylesheet(app, "dark_teal.xml", invert_secondary=False, extra=extra, save_as="stylesheet.css")

    app.startup()

    app.exec()



from PySide6 import QtWidgets, QtCore
import qtawesome as qta


class QHLine(QtWidgets.QFrame):
    def __init__(self):
        super(QHLine, self).__init__()
        self.setFrameShape(QtWidgets.QFrame.HLine)
        self.setFrameShadow(QtWidgets.QFrame.Sunken)


class QVLine(QtWidgets.QFrame):
    def __init__(self):
        super(QVLine, self).__init__()
        self.setFrameShape(QtWidgets.QFrame.VLine)
        self.setFrameShadow(QtWidgets.QFrame.Sunken)


class DockTitleBar(QtWidgets.QWidget):
    def __init__(self, title, clicked=None, parent=None):
        super().__init__(parent)
        self.app = QtWidgets.QApplication.instance()

        self.layout = QtWidgets.QHBoxLayout()
        self.setLayout(self.layout)
        self.layout.setContentsMargins(3, 3, 3, 3)
        self.layout.setSpacing(3)

        icon = QtWidgets.QLabel()
        icon.setPixmap(qta.icon("fa5s.grip-vertical", color=self.app.theme.icon_color).pixmap(15))
        self.layout.addWidget(icon)

        label = QtWidgets.QLabel(title)
        label.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        self.layout.addWidget(label)

        close = QtWidgets.QPushButton("", icon=qta.icon("fa5s.times", color=self.app.theme.icon_color), objectName="close", clicked=clicked)
        close.setFlat(True)
        close.setMouseTracking(True)
        self.layout.addWidget(close)
        self.setAttribute(QtCore.Qt.WidgetAttribute.WA_StyledBackground, True)

        self.setStyleSheet(
            f"""
            DockTitleBar {{
                background:rgba(0,0,0,0.1);
            }}

            QLabel {{
                background: transparent;
            }}

            QPushButton {{
                background: rgba(0, 0, 0, 0.1);
                border: 0px solid white;
                border-radius: 2px;
            }}

            QPushButton::hover {{
                background: rgba(0, 0, 0, 0.3);
            }}
            """
        )


class IconLabel(QtWidgets.QWidget):

    HorizontalSpacing = 2

    def __init__(self, qta_id, text="", final_stretch=True, icon_size=16):
        super().__init__()

        self.app = QtWidgets.QApplication.instance()

        self.icon_size = QtCore.QSize(icon_size, icon_size)

        layout = QtWidgets.QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)

        self.icon = QtWidgets.QLabel()
        self.text = QtWidgets.QLabel()

        self.setIcon(qta_id)
        self.setText(text)

        layout.addWidget(self.icon)
        layout.addSpacing(self.HorizontalSpacing)
        layout.addWidget(self.text)

        if final_stretch:
            layout.addStretch()

    def setText(self, text):
        self.text.setText(text)

    def setIcon(self, qta_id):
        self.icon.setPixmap(qta.icon(qta_id, color=self.app.theme.icon_color).pixmap(self.icon_size))



from PySide6 import QtWidgets, QtCore, QtGui
from meni.ui.common import DockTitleBar
from enum import IntFlag, auto
import qtawesome as qta


class DeselectableTreeView(QtWidgets.QTreeView):
    def mousePressEvent(self, event):

        if event.button() == QtCore.Qt.MouseButton.LeftButton:
            index = self.indexAt(event.pos())
            if not index.isValid():
                self.clearSelection()

        super().mousePressEvent(event)


class BrowserDock(QtWidgets.QDockWidget):
    def __init__(self, parent):
        super().__init__("Browser", objectName="browser", parent=parent)
        self.setStyleSheet("background: transparent;")

        self.app = QtCore.QCoreApplication.instance()
        self.app.add_filter(self.filter)

        self.setTitleBarWidget(DockTitleBar("Browser", clicked=self.close))

        self.setFeatures(QtWidgets.QDockWidget.DockWidgetFeature.DockWidgetMovable | QtWidgets.QDockWidget.DockWidgetFeature.DockWidgetClosable)
        self.setAllowedAreas(QtCore.Qt.DockWidgetArea.AllDockWidgetAreas)

        self.tree = DeselectableTreeView()
        self.tree.setIndentation(10)
        self.tree.setRootIsDecorated(True)
        self.tree.setHeaderHidden(True)
        self.model = BrowserModel()
        self.tree.setModel(self.model)
        self.tree.setFocusPolicy(QtCore.Qt.FocusPolicy.NoFocus)

        self.tree.header().setStretchLastSection(False)
        self.tree.header().setSectionResizeMode(0, QtWidgets.QHeaderView.ResizeMode.Stretch)
        self.tree.header().setSectionResizeMode(1, QtWidgets.QHeaderView.ResizeMode.ResizeToContents)

        self.setWidget(self.tree)

        self.tree.setStyleSheet(
            f"""
                QTreeView::item {{
                    padding: 3px 0px;
                }}
            """
        )

        self.app.metadata.changed.connect(self.model.refresh)
        self.tree.clicked.connect(self.on_click)

    def on_click(self, index):
        item = self.model.get_item(index)
        if TreeItemType.ROOT_FOLDER not in item.type:
            if self.model.highlighted == item:
                self.model.highlighted = None
            else:
                self.model.highlighted = item
            self.app.filter_changed.emit()
        else:
            if self.tree.isExpanded(index):
                self.tree.collapse(index)
            else:
                self.tree.expand(index)

    def filter(self, file):
        highlighted = self.model.highlighted
        if highlighted:
            if TreeItemType.TAG in highlighted.type:
                if TreeItemType.NO_VALUE in highlighted.type:
                    return len(file.tags) == 0
                for tag in file.tags:
                    if tag.startswith(highlighted.data):
                        return True
            if TreeItemType.COLLECTION in highlighted.type:
                if TreeItemType.NO_VALUE in highlighted.type:
                    return file.collection is None or file.collection.strip() == ""
                return file.collection and file.collection.startswith(highlighted.data)
            if TreeItemType.AUTHOR in highlighted.type:
                if TreeItemType.NO_VALUE in highlighted.type:
                    return file.collection_obj is None or not file.collection_obj.author
                return file.collection_obj and file.collection_obj.author == highlighted.data
            return False

        return True


class BrowserModel(QtCore.QAbstractItemModel):
    def __init__(self):
        super().__init__()
        self.app = QtCore.QCoreApplication.instance()
        self._highlighted = None
        self.root_item = BrowserTreeItem("root", None, TreeItemType.ROOT_FOLDER)

        self.tags_item = self.root_item.append_child("Tags", TreeItemType.ROOT_FOLDER | TreeItemType.TAG)
        self.collections_item = self.root_item.append_child("Collections", TreeItemType.ROOT_FOLDER | TreeItemType.COLLECTION)
        self.authors_item = self.root_item.append_child("Authors", TreeItemType.ROOT_FOLDER | TreeItemType.AUTHOR)

        self.notags_item = self.tags_item.append_child("No Tags", TreeItemType.NO_VALUE | TreeItemType.TAG)
        self.nocollections_item = self.collections_item.append_child("No Collection", TreeItemType.NO_VALUE | TreeItemType.COLLECTION)
        self.noauthor_item = self.authors_item.append_child("No Author", TreeItemType.NO_VALUE | TreeItemType.AUTHOR)

        self.refresh()

    def data(self, index: QtCore.QModelIndex, role: int):
        if not index.isValid():
            return None

        item: BrowserTreeItem = self.get_item(index)

        if role == QtCore.Qt.ItemDataRole.DisplayRole:
            if index.column() == 0:
                text = item.text
                if item == self.highlighted:
                    text = f"{text} (filtered)"
                if TreeItemType.ROOT_FOLDER in item.type:
                    text = f"[ {text.upper()} ]"
                return text
            elif index.column() == 1:
                if TreeItemType.ROOT_FOLDER not in item.type:
                    return item.count

        elif role == QtCore.Qt.ItemDataRole.DecorationRole and index.column() == 0:

            if item == self.highlighted:
                icon = "fa5s.filter"
            elif TreeItemType.ROOT_FOLDER | TreeItemType.TAG in item.type:
                icon = "fa5s.tags"
            elif TreeItemType.ROOT_FOLDER | TreeItemType.COLLECTION in item.type:
                icon = "fa5s.layer-group"
            elif TreeItemType.ROOT_FOLDER | TreeItemType.AUTHOR in item.type:
                icon = "fa5s.user-friends"
            elif TreeItemType.NO_VALUE in item.type:
                icon = "fa5s.minus-circle"
            elif TreeItemType.TAG in item.type:
                icon = "fa5s.tag"
            elif TreeItemType.COLLECTION in item.type:
                icon = "fa5s.th-large"
            elif TreeItemType.AUTHOR in item.type:
                icon = "fa5s.user"

            color = self.app.theme.icon_color
            if self.highlighted:
                if self.highlighted != item and not self.highlighted.path.startswith(item.path):
                    color = self.app.theme.muted

            return qta.icon(icon, color=color)

        elif role == QtCore.Qt.ItemDataRole.TextAlignmentRole and index.column() == 1:
            return QtCore.Qt.AlignmentFlag.AlignRight | QtCore.Qt.AlignmentFlag.AlignVCenter

        elif role == QtCore.Qt.ItemDataRole.FontRole and (
            item == self.highlighted or (self.highlighted and self.highlighted.path.startswith(item.path))
        ):
            font = self.app.font()
            font.setBold(True)
            return font

        elif role == QtCore.Qt.ItemDataRole.ForegroundRole:
            if self.highlighted:
                if self.highlighted != item and not self.highlighted.path.startswith(item.path):
                    return QtGui.QColor(self.app.theme.muted)
            if TreeItemType.ROOT_FOLDER in item.type:
                return QtGui.QColor(self.app.theme.icon_color)

        return None

    def index(self, row: int, column: int, parent: QtCore.QModelIndex = QtCore.QModelIndex()):
        parent_item: BrowserTreeItem = self.get_item(parent)

        if not parent_item:
            return QtCore.QModelIndex()

        child_item: BrowserTreeItem = parent_item.child(row)
        if child_item:
            return self.createIndex(row, column, child_item)
        return QtCore.QModelIndex()

    def parent(self, index: QtCore.QModelIndex):
        if not index.isValid():
            return QtCore.QModelIndex()

        child_item: BrowserTreeItem = self.get_item(index)
        if child_item:
            parent_item: BrowserTreeItem = child_item.parent()
        else:
            parent_item = None

        if parent_item == self.root_item or not parent_item:
            return QtCore.QModelIndex()
        return self.createIndex(parent_item.child_number(), 0, parent_item)

    def columnCount(self, parent: QtCore.QModelIndex):
        return 2

    def rowCount(self, parent: QtCore.QModelIndex):
        parent_item: BrowserTreeItem = self.get_item(parent)
        if parent_item:
            return len(parent_item.children)
        return 0

    def get_item(self, index: QtCore.QModelIndex):
        if index.isValid():
            item = index.internalPointer()
            if item:
                return item
        return self.root_item

    def get_item_by_path(self, path):
        return self.root_item.child_by_path(path)

    def refresh(self):
        metadata = QtCore.QCoreApplication.instance().metadata

        self.layoutAboutToBeChanged.emit()
        self.root_item.reset()

        for file in metadata.files:

            tags = file.tags
            if len(tags) == 0:
                self.notags_item.count += 1
            else:
                for tag in tags:
                    tag_path = ["Tags"]
                    for tag_part in tag.split("/"):
                        tag_path = tag_path + [tag_part]
                        tag_item = self.get_item_by_path(tag_path)
                        if tag_item is None:
                            tag_item = self.get_item_by_path(tag_path[:-1]).append_child("/".join(tag_path[1:]), TreeItemType.TAG)
                            tag_item.count = 0
                        tag_item.count += 1

            coll = file.collection
            if coll is None or coll.strip() == "":
                self.nocollections_item.count += 1
            else:
                coll_path = ["Collections"]
                for coll_part in coll.split("/"):
                    coll_path = coll_path + [coll_part]
                    coll_item = self.get_item_by_path(coll_path)
                    if coll_item is None:
                        coll_item = self.get_item_by_path(coll_path[:-1]).append_child("/".join(coll_path[1:]), TreeItemType.COLLECTION)
                        coll_item.count = 0
                    coll_item.count += 1

            author = file.collection_obj.author if file.collection_obj and file.collection_obj.author else None
            if author is None or author.strip() == "":
                self.noauthor_item.count += 1
            else:
                author_item = self.get_item_by_path(["Authors", author])
                if author_item is None:
                    author_item = self.get_item_by_path(["Authors"]).append_child(author, TreeItemType.AUTHOR)
                    author_item.count = 0
                author_item.count += 1

        self.layoutChanged.emit()

    @property
    def highlighted(self):
        return self._highlighted

    @highlighted.setter
    def highlighted(self, item):
        self.layoutAboutToBeChanged.emit()
        self._highlighted = item
        self.layoutChanged.emit()

    def flags(self, index: QtCore.QModelIndex | QtCore.QPersistentModelIndex) -> QtCore.Qt.ItemFlag:
        return QtCore.Qt.ItemFlag.ItemIsEnabled


class TreeItemType(IntFlag):
    ROOT_FOLDER = auto()
    TAG = auto()
    COLLECTION = auto()
    AUTHOR = auto()
    NO_VALUE = auto()


class BrowserTreeItem:
    def __init__(self, data, parent: "BrowserTreeItem", type: TreeItemType):
        self.data = data
        self.parent_item = parent
        self.children = []
        self.count = 0
        self.type = type

    def child(self, row: int) -> "BrowserTreeItem":
        if row >= len(self.children) or row < 0:
            return None
        return self.children[row]

    def child_by_path(self, path):
        for item in self.children:
            if item.text == path[0]:
                if len(path) == 1:
                    return item
                return item.child_by_path(path[1:])
        return None

    @property
    def text(self):
        return self.data.split("/")[-1]

    def child_number(self) -> int:
        if self.parent_item:
            return self.parent_item.children.index(self)
        return 0

    def append_child(self, data, type: TreeItemType) -> "BrowserTreeItem":
        child = BrowserTreeItem(data, self, type)
        self.children.append(child)
        return child

    @property
    def path(self):
        if self.parent():
            return f"{self.parent().path}/{self.text}"
        else:
            return self.text

    def reset(self):
        self.count = 0
        for child in self.children:
            child.reset()

    def parent(self) -> "BrowserTreeItem":
        return self.parent_item

    def __repr__(self):
        return f"BrowserTreeItem({self.type}, {self.path})"



from PySide6 import QtWidgets, QtCore, QtGui
from meni.ui.searchinput import SearchInput
from meni.ui.windows.importdialog import ImportDialog
from meni.ui.menus.menusettings import MenuSettings

import qtawesome as qta


class MainToolbar(QtWidgets.QToolBar):
    def __init__(self, parent=None):
        super().__init__("Main", parent=parent)

        self.app = QtCore.QCoreApplication.instance()

        self.setObjectName("main_toolbar")
        self.setToolButtonStyle(QtCore.Qt.ToolButtonStyle.ToolButtonTextBesideIcon)
        self.setMovable(False)
        self.setFloatable(False)
        self.setIconSize(QtCore.QSize(16, 16))

        import_action = QtGui.QAction("Import", self, icon=qta.icon("fa5.plus-square", color=self.app.theme.icon_color), text="Import")
        import_action.setShortcut("Ctrl+I")
        import_action.triggered.connect(lambda: ImportDialog(self).exec())
        self.addAction(import_action)

        self.addSeparator()

        self.addWidget(SearchInput())

        self.addSeparator()
        settings_action = QtGui.QAction("Settings", self, icon=qta.icon("fa5s.cog", color=self.app.theme.icon_color), text="Settings")
        settings_action.setMenu(MenuSettings(self))
        self.addAction(settings_action)



from PySide6 import QtWidgets, QtCore, QtGui
import qtawesome as qta


class MenuSettings(QtWidgets.QMenu):
    def __init__(self, parent):
        super().__init__(parent)
        self.app = QtCore.QCoreApplication.instance()
        open_library_folder_action = QtGui.QAction(
            f"Open Library Folder...", self, icon=qta.icon("fa5s.folder-open", color=self.app.theme.icon_color), text="Open Library Folder..."
        )
        open_library_folder_action.triggered.connect(self.on_open_library_folder)
        self.addAction(open_library_folder_action)

    def on_open_library_folder(self):
        folder = QtWidgets.QFileDialog.getExistingDirectory(None, "Select a folder")
        if folder:
            self.app.current_library = folder
            self.app.metadata.changed.emit()



from PySide6 import QtWidgets, QtCore, QtGui
from meni.ui.menus.filecontextmenu import FileContextMenu
from meni.ui.windows.importdialog import ImportDialog
from meni.utils import tags_from_text
import qtawesome as qta


class FilesTable(QtWidgets.QTableView):
    def __init__(self):
        super().__init__()
        self.app = QtCore.QCoreApplication.instance()
        self.delegate = PreviewDelegate()

        self.setSelectionMode(QtWidgets.QAbstractItemView.SelectionMode.ExtendedSelection)
        self.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
        self.setModel(TableModel())
        self.setFocusPolicy(QtCore.Qt.FocusPolicy.NoFocus)

        self.setItemDelegateForColumn(0, self.delegate)
        self.verticalHeader().setDefaultSectionSize(self.delegate.size)
        self.verticalHeader().hide()

        self.selectionModel().selectionChanged.connect(self.on_selection_changed)
        header = self.horizontalHeader()

        header = self.horizontalHeader()
        header.setSectionResizeMode(QtWidgets.QHeaderView.ResizeToContents)

        self.setDragEnabled(True)
        self.setDragDropMode(QtWidgets.QAbstractItemView.DragDropMode.DragDrop)
        self.setDragDropOverwriteMode(False)
        self.setAcceptDrops(True)

        self.app.filter_changed.connect(self.on_filter_changed)
        self.app.metadata.changed.connect(self.model().layoutChanged.emit)

    def contextMenuEvent(self, event):
        files = [self.model().files[row.row()] for row in self.selectionModel().selectedRows()]
        if files:
            menu = FileContextMenu(self, files)
            menu.popup(event.globalPos())

        return super().contextMenuEvent(event)

    def on_filter_changed(self):
        self.selectionModel().clearSelection()
        self.model().layoutChanged.emit()

    def on_selection_changed(self):
        app = QtCore.QCoreApplication.instance()

        if self.selectionModel().hasSelection():
            app.selected_files = [self.model().files[row.row()] for row in self.selectionModel().selectedRows()]
        else:
            app.selected_files = None

    def dragEnterEvent(self, event):
        if event.mimeData().hasUrls():
            event.accept()
        else:
            event.ignore()

    def dragMoveEvent(self, event: QtGui.QDragMoveEvent) -> None:
        if event.mimeData().hasUrls():
            event.accept()
        else:
            event.ignore()

    def dropEvent(self, event):

        files = []
        for url in event.mimeData().urls():
            files.append(url.toLocalFile())
        event.accept()
        ImportDialog(self, files).exec_()

    @property
    def size(self):
        return self.delegate.size

    @size.setter
    def size(self, value):
        self.delegate.size = value
        self.verticalHeader().setDefaultSectionSize(value)
        self.model().layoutChanged.emit()


class PreviewDelegate(QtWidgets.QStyledItemDelegate):

    def __init__(self):
        super().__init__()
        self.size = 50

    def paint(self, painter, option, index):

        # data is our preview object
        data = index.model().data(index, QtCore.Qt.DisplayRole)

        if data is None:
            return

        image = QtGui.QImage(data)

        scaled = image.scaled(
            self.size,
            self.size,
            QtCore.Qt.AspectRatioMode.KeepAspectRatio,
        )
        # Position in the middle of the area.
        # x = CELL_PADDING + (width - scaled.width()) / 2
        # y = CELL_PADDING + (height - scaled.height()) / 2

        painter.drawImage(option.rect.x(), option.rect.y(), scaled)

    def sizeHint(self, option, index):
        # All items the same size.
        return QtCore.QSize(self.size, self.size)


THUMBNAIL, NAME, COLLECTION, TAGS, PATH = range(5)


class TableModel(QtCore.QAbstractTableModel):

    def __init__(self):
        super(TableModel, self).__init__()
        self._files = QtCore.QCoreApplication.instance().metadata.files
        self.app = QtCore.QCoreApplication.instance()

    @property
    def files(self):
        app = QtCore.QCoreApplication.instance()

        filtered_files = self._files

        for filter in app.filters:
            filtered_files = [file for file in filtered_files if filter(file)]

        if app._search_filter:
            filtered_files = [file for file in filtered_files if app._search_filter.lower() in file.name.lower()]

        return filtered_files

    def headerData(self, section, orientation, role):
        if role == QtCore.Qt.ItemDataRole.DisplayRole:
            if orientation == QtCore.Qt.Orientation.Horizontal:
                if section == THUMBNAIL:
                    return ""
                elif section == NAME:
                    return "Name"
                elif section == COLLECTION:
                    return "Collection"
                elif section == TAGS:
                    return "Tags"
                elif section == PATH:
                    return "Path"
            else:
                return section + 1
        return None

    def rowCount(self, index):
        return len(self.files)

    def columnCount(self, index):
        return 5

    def data(self, index, role):
        if role == QtCore.Qt.ItemDataRole.DisplayRole or role == QtCore.Qt.ItemDataRole.EditRole:
            if index.column() == THUMBNAIL:
                return self.files[index.row()].thumbnail_file
            elif index.column() == NAME:
                return self.files[index.row()].name
            elif index.column() == COLLECTION:
                return self.files[index.row()].collection
            elif index.column() == TAGS:
                return ", ".join(self.files[index.row()].tags)
            elif index.column() == PATH:
                return self.files[index.row()].path

        elif role == QtCore.Qt.DecorationRole:
            if index.column() == COLLECTION:
                if self.files[index.row()].collection_obj is not None:
                    return qta.icon("fa5s.layer-group", color=self.app.theme.icon_color)
            elif index.column() == TAGS:
                if self.files[index.row()].tags:
                    return qta.icon("fa5s.tag", color=self.app.theme.icon_color)
            elif index.column() == PATH:
                return qta.icon("fa5s.file", color=self.app.theme.icon_color)

    def flags(self, index):
        flags = QtCore.Qt.ItemIsSelectable | QtCore.Qt.ItemIsEnabled | QtCore.Qt.ItemIsDragEnabled
        if index.column() == NAME or index.column() == COLLECTION or index.column() == TAGS:
            flags |= QtCore.Qt.ItemIsEditable
        return flags

    def validate(self, index, value):
        if index.column() == NAME:
            return value.strip() != ""
        if index.column() == COLLECTION:
            return True
        if index.column() == TAGS:
            return True
        return False

    def mimeData(self, indexes):

        rows = set(index.row() for index in indexes)
        urls = [QtCore.QUrl.fromLocalFile(self.files[row].absolute_path) for row in rows]

        mime_data = QtCore.QMimeData()
        mime_data.setUrls(urls)
        return mime_data

    def setData(self, index, value, role):
        if role == QtCore.Qt.EditRole:
            app = QtCore.QCoreApplication.instance()
            if index.column() == NAME:
                if not self.validate(index, value):
                    return False
                app.metadata.update_file(self.files[index.row()], name=value)
            if index.column() == COLLECTION:
                app.metadata.update_file(self.files[index.row()], collection=value)
            if index.column() == TAGS:
                app.metadata.update_file(self.files[index.row()], tags=tags_from_text(value))
            return True



from PySide6 import QtWidgets, QtCore, QtGui
from meni.ui.windows.mainwindow import MainWindow
from meni.utils import pxmap_from_svg
import meni.rc_assets


class WelcomeWindow(QtWidgets.QWidget):

    def __init__(self):
        super().__init__()

        # Attributes
        self.library_path = None

        self.app = QtCore.QCoreApplication.instance()

        # UI Components
        self.setStyleSheet(
            f"""
                           background-color: {self.app.theme.main_background}; 
                           color: {self.app.theme.main_foreground};
                           selection-background-color: {self.app.theme.selection_background};
                           selection-color: {self.app.theme.selection_foreground};
                           """
        )

        self.setWindowTitle("Welcome to Meni 3D Library")

        self.layout = QtWidgets.QVBoxLayout(self)
        self.layout.setContentsMargins(30, 15, 30, 15)

        # Logo (logo.svg)
        self.logo = QtWidgets.QLabel()
        self.logo.setPixmap(
            pxmap_from_svg(":/assets/logo.svg", color=self.app.theme.icon_color).scaled(
                120, 120, QtCore.Qt.KeepAspectRatio, QtCore.Qt.SmoothTransformation
            )
        )

        self.logo.setAlignment(QtCore.Qt.AlignCenter)
        self.layout.addWidget(self.logo, 0)

        # Panel
        self.panel = QtWidgets.QFrame()
        self.panel.setFrameShape(QtWidgets.QFrame.Box)
        self.panel.setFrameShadow(QtWidgets.QFrame.Raised)
        self.panel.setStyleSheet(
            """
                QFrame {
                        background-color: rgba(255, 255, 255, 0.01);
                        }            
                QLabel {
                    background: transparent;
                    border: none;
                    }
                                 """
        )
        self.panel.layout = QtWidgets.QVBoxLayout(self.panel)
        self.panel.layout.setContentsMargins(30, 15, 30, 30)
        self.panel.layout.setSpacing(15)
        self.layout.addWidget(self.panel, 1)

        # Text description
        self.description = QtWidgets.QLabel(
            """<strong>Welcome to Meni 3D Library</strong>
                <br><br>
                To get started with Meni, you'll need to designate a folder as your library. This folder will house your imported 3D models along with their associated metadata, including tags.
                <br><br>
                Please avoid selecting a folder that already contains your models; instead, import them into the library directly through the application.
                                    """,
        )
        self.description.setWordWrap(True)
        self.panel.layout.addWidget(self.description)

        # Path selection
        self.path = QtWidgets.QLabel(QtCore.QCoreApplication.instance().current_library)
        self.path.setStyleSheet(
            "border: 1px solid rgba(255,255,255, 0.2); padding: 5px; border-radius: 5px; background-color: rgba(255,255,255, 0.1); "
        )
        self.btn_browse = QtWidgets.QPushButton("Browse")
        self.btn_browse.clicked.connect(self.select_folder)

        self.rowLine = QtWidgets.QHBoxLayout()
        self.rowLine.addWidget(self.path, 1)
        self.rowLine.addWidget(self.btn_browse, 0)
        self.panel.layout.addLayout(self.rowLine)

        # Bottom buttons
        self.bottomButtons = QtWidgets.QHBoxLayout()
        self.bottomButtons.setContentsMargins(0, 15, 0, 0)
        self.layout.addLayout(self.bottomButtons, 0)

        self.btn_quit = QtWidgets.QPushButton("Quit")
        self.btn_quit.clicked.connect(self.close)
        self.bottomButtons.addWidget(self.btn_quit, 0)

        self.bottomButtons.addSpacerItem(QtWidgets.QSpacerItem(0, 0, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum))

        self.btn_init = QtWidgets.QPushButton("Initialize Library")
        self.btn_init.setEnabled(QtCore.QCoreApplication.instance().current_library is not None)
        self.btn_init.setMinimumWidth(180)
        self.btn_init.setDefault(True)
        self.btn_init.clicked.connect(self.init)
        self.bottomButtons.addWidget(self.btn_init, 0)

    def select_folder(self):
        folder = QtWidgets.QFileDialog.getExistingDirectory(None, "Select a folder")
        if folder:
            self.library_path = folder
            self.path.setText(folder)
            self.btn_init.setEnabled(True)

    def init(self):
        app = QtCore.QCoreApplication.instance()
        app.current_library = self.library_path
        app.show_main()
        self.hide()

    def close(self):
        QtCore.QCoreApplication.quit()



from PySide6 import QtWidgets, QtCore, QtGui
from meni.ui.common import DockTitleBar, IconLabel
from meni.utils import tags_from_text

import qtawesome as qta


class PropertyRowWithApply(QtWidgets.QWidget):
    def __init__(self, parent, widget, apply_dict_fn, has_apply_all=True):
        super().__init__(parent)

        self.app = QtCore.QCoreApplication.instance()
        self.has_apply_all = has_apply_all
        self.apply_dict_fn = apply_dict_fn

        self.layout = QtWidgets.QHBoxLayout()
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.layout.setSpacing(2)
        self.setLayout(self.layout)

        self.layout.addWidget(widget)

        self.btn_apply = QtWidgets.QPushButton("", objectName="apply")
        self.btn_apply.setIcon(qta.icon("fa5s.check", color=self.app.theme.icon_color))
        self.btn_apply.clicked.connect(self.apply_last)
        self.btn_apply.setSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        self.btn_apply.setToolTip("<html>Apply to <b>last</b> selected file</html>")
        self.layout.addWidget(self.btn_apply)

        if self.has_apply_all:
            self.btn_apply_all = QtWidgets.QPushButton("", objectName="apply_all")
            self.btn_apply_all.setIcon(qta.icon("fa5s.check-double", color=self.app.theme.icon_color))
            self.btn_apply_all.clicked.connect(self.apply_all)
            self.btn_apply_all.setSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
            self.btn_apply_all.setToolTip("<html>Apply to <b>all</b> selected files</html>")
            self.layout.addWidget(self.btn_apply_all)

        self.app.selected_files_changed.connect(self.on_selected_file_changed)

    def apply_last(self):
        self.apply(self.app.last_selected_file, **self.apply_dict_fn())

    def apply_all(self):
        args = self.apply_dict_fn()
        for file in self.app.selected_files:
            self.apply(file, **args)

    def apply(self, file, **kwargs):
        pass

    def on_selected_file_changed(self, files, last):
        if last:
            if self.has_apply_all:
                self.btn_apply_all.setVisible(len(files) > 1)
                self.btn_apply_all.setToolTip(f"<html>Apply to <b>all ({len(files)})</b> selected files</html>")

            self.btn_apply.setToolTip(f"<html>Apply to last selected file, <b>{last}</b> (ENTER)</html>")


class FilePropertyRowWithApply(PropertyRowWithApply):
    def __init__(self, parent, widget, apply_dict_fn, has_apply_all=True):
        super().__init__(parent, widget, apply_dict_fn, has_apply_all)

        self.apply_dict_fn = apply_dict_fn

    def apply(self, file, **kwargs):
        if file:
            self.app.metadata.update_file(file, **kwargs)


class CollectionPropertyRowWithApply(PropertyRowWithApply):
    def __init__(self, parent, widget, apply_dict_fn, has_apply_all=True):
        super().__init__(parent, widget, apply_dict_fn, has_apply_all)

        self.apply_dict_fn = apply_dict_fn

    def apply(self, file, **kwargs):
        if file and file.collection_obj:
            self.app.metadata.update_collection(file.collection_obj, **kwargs)



from PySide6 import QtCore


class Worker(QtCore.QRunnable):
    """
    Worker thread

    Inherits from QRunnable to handler worker thread setup, signals and wrap-up.

    :param callback: The function callback to run on this worker thread. Supplied args and
                     kwargs will be passed through to the runner.
    :type callback: function
    :param args: Arguments to pass to the callback function
    :param kwargs: Keywords to pass to the callback function

    """

    def __init__(self, fn, *args, **kwargs):
        super(Worker, self).__init__()
        # Store constructor arguments (re-used for processing)
        self.fn = fn
        self.args = args
        self.kwargs = kwargs

    @QtCore.Slot()
    def run(self):
        """
        Initialise the runner function with passed args, kwargs.
        """
        self.fn(*self.args, **self.kwargs)



from PySide6 import QtWidgets, QtCore, QtGui
from meni.ui.common import DockTitleBar, IconLabel
from meni.ui.propertyrowwithapply import FilePropertyRowWithApply
from meni.utils import tags_from_text

import qtawesome as qta


class FilePropertiesDock(QtWidgets.QDockWidget):
    def __init__(self, parent):
        super().__init__("File Properties", objectName="properties", parent=parent)

        self.setTitleBarWidget(DockTitleBar("File Properties", clicked=self.close))

        self.app = QtCore.QCoreApplication.instance()

        self.layout = QtWidgets.QFormLayout()

        self.empty = QtWidgets.QLabel("No file selected")
        self.empty.setAlignment(QtCore.Qt.AlignCenter)
        self.empty.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        self.empty.setStyleSheet("color: rgba(255,255,255,0.2);")
        self.layout.addRow(self.empty)

        # === Name
        self.name = QtWidgets.QLineEdit()
        self.name.returnPressed.connect(lambda: self.apply_file(name=self.name.text()))
        self.layout.addRow(
            "Name",
            FilePropertyRowWithApply(
                self,
                self.name,
                lambda: dict(name=self.name.text()),
                has_apply_all=False,
            ),
        )

        # === Collection
        self.collection = QtWidgets.QComboBox(self)
        self.collection.addItems([col.name for col in self.app.metadata.collections])
        self.collection_edit = QtWidgets.QLineEdit(self)
        self.collection.setLineEdit(self.collection_edit)
        self.collection_edit.setText("")
        self.collection_edit.returnPressed.connect(lambda: self.apply_file(collection=self.collection_edit.text()))

        self.layout.addRow(
            "Collection",
            FilePropertyRowWithApply(
                self,
                self.collection,
                lambda: dict(collection=self.collection_edit.text()),
            ),
        )

        # === Tags
        self.tags = QtWidgets.QLineEdit()
        self.tags.returnPressed.connect(lambda: self.apply_file(tags=tags_from_text(self.tags.text())))
        self.layout.addRow(
            "Tags",
            FilePropertyRowWithApply(
                self,
                self.tags,
                lambda: dict(tags=tags_from_text(self.tags.text())),
            ),
        )

        # Set layout to widget to dock
        widget = QtWidgets.QWidget()
        widget.setLayout(self.layout)
        self.setWidget(widget)

        # Connect signals
        self.app.metadata.changed.connect(lambda: self.refresh([], self.app.last_selected_file))
        self.app.selected_files_changed.connect(self.refresh)

        self.refresh([], None)

    def refresh(self, files, last):
        self.layout.setRowVisible(0, last is None)
        for i in range(1, self.layout.rowCount()):
            self.layout.setRowVisible(i, last is not None)

        self.name.setText(last.name if last else "")
        self.collection_edit.setText(last.collection if last else "")
        self.tags.setText(", ".join(last.tags) if last else "")

    def apply_file(self, **kwargs):
        if self.app.last_selected_file:
            self.app.metadata.update_file(self.app.last_selected_file, **kwargs)



class Gruvbox:
    def __init__(self):
        self.background = "#282828"
        self.red = "#cc241d"
        self.green = "#98971a"
        self.yellow = "#d79921"
        self.blue = "#458588"
        self.purple = "#b16286"
        self.aqua = "#689d6a"
        self.gray = "#a89984"
        self.orange = "#d65d0e"
        self.bright_red = "#fb4934"
        self.bright_green = "#b8bb26"
        self.bright_yellow = "#fabd2f"
        self.bright_blue = "#83a598"
        self.bright_purple = "#d3869b"
        self.bright_aqua = "#8ec07c"
        self.bright_orange = "#fe8019"

        self.foreground = "#ebdbb2"
        self.foreground0 = "#fbf1c7"
        self.foreground1 = "#ebdbb2"
        self.foreground2 = "#d5c4a1"
        self.foreground3 = "#bdae93"
        self.foreground4 = "#a89984"

        self.background0 = "#282828"
        self.background1 = "#3c3836"
        self.background2 = "#504945"
        self.background3 = "#665c54"
        self.background4 = "#7c6f64"

    @property
    def main_background(self):
        return self.background

    @property
    def main_foreground(self):
        return self.foreground

    @property
    def icon_color(self):
        return self.bright_blue

    @property
    def model_color(self):
        return self.aqua

    @property
    def tag_background(self):
        return self.gray

    @property
    def tag_foreground(self):
        return self.background

    @property
    def selection_background(self):
        return self.background4

    @property
    def selection_foreground(self):
        return self.background

    @property
    def green_btn(self):
        return self.green

    @property
    def red_btn(self):
        return self.red

    @property
    def muted(self):
        return "#666666"


class Nord:
    def __init__(self):

        self.background = "#2e3440"
        self.foreground = "#d8dee9"
        self.comment = "#4c566a"
        self.red = "#bf616a"
        self.green = "#a3be8c"
        self.yellow = "#ebcb8b"
        self.blue = "#81a1c1"
        self.purple = "#b48ead"
        self.aqua = "#88c0d0"
        self.orange = "#d08770"
        self.bright_red = "#bf616a"
        self.bright_green = "#a3be8c"
        self.bright_yellow = "#ebcb8b"
        self.bright_blue = "#81a1c1"
        self.bright_purple = "#b48ead"
        self.bright_aqua = "#88c0d0"
        self.bright_orange = "#d08770"

        self.foreground0 = "#d8dee9"
        self.foreground1 = "#e5e9f0"
        self.foreground2 = "#eceff4"
        self.foreground3 = "#8fbcbb"
        self.foreground4 = "#88c0d0"

        self.background0 = "#2e3440"
        self.background1 = "#3b4252"
        self.background2 = "#434c5e"
        self.background3 = "#4c566a"
        self.background4 = "#5e81ac"

    @property
    def main_background(self):
        return self.background

    @property
    def main_foreground(self):
        return self.foreground

    @property
    def icon_color(self):
        return self.bright_blue

    @property
    def model_color(self):
        return self.green

    @property
    def tag_background(self):
        return self.orange

    @property
    def tag_foreground(self):
        return self.background0

    @property
    def selection_background(self):
        return self.background4

    @property
    def selection_foreground(self):
        return self.background

    @property
    def green_btn(self):
        return self.green

    @property
    def red_btn(self):
        return self.red

    @property
    def muted(self):
        return "#5b6880"


class Dracula:
    def __init__(self):
        self.background = "#282a36"
        self.current_line = "#44475a"
        self.foreground = "#f8f8f2"
        self.comment = "#6272a4"
        self.cyan = "#8be9fd"
        self.green = "#50fa7b"
        self.orange = "#ffb86c"
        self.pink = "#ff79c6"
        self.purple = "#bd93f9"
        self.red = "#ff5555"
        self.yellow = "#f1fa8c"

    @property
    def main_background(self):
        return self.background

    @property
    def main_foreground(self):
        return self.foreground

    @property
    def icon_color(self):
        return self.cyan

    @property
    def model_color(self):
        return self.green

    @property
    def tag_background(self):
        return self.red

    @property
    def tag_foreground(self):
        return self.background

    @property
    def selection_background(self):
        return self.comment

    @property
    def selection_foreground(self):
        return self.foreground

    @property
    def green_btn(self):
        return self.green

    @property
    def red_btn(self):
        return self.red

    @property
    def muted(self):
        return "#5f6380"



from PySide6 import QtWidgets, QtCore, QtGui
import qtawesome as qta


class SearchInput(QtWidgets.QWidget):
    def __init__(self):
        super().__init__()

        self.app = QtCore.QCoreApplication.instance()

        self.layout = QtWidgets.QHBoxLayout()
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(self.layout)

        self.search = QtWidgets.QLineEdit()
        self.search.setPlaceholderText("Search")
        search_action = QtGui.QAction("Search", self, icon=qta.icon("fa5s.search", color=self.app.theme.icon_color, text=""))
        self.search.addAction(search_action, QtWidgets.QLineEdit.LeadingPosition)
        self.layout.addWidget(self.search)

        self.clear_search = QtGui.QAction("Clear", self, icon=qta.icon("fa5s.backspace", color=self.app.theme.icon_color, text=""))
        self.clear_search.triggered.connect(lambda: self.search.clear())
        self.clear_search.setEnabled(False)
        self.search.addAction(self.clear_search, QtWidgets.QLineEdit.TrailingPosition)

        self.search.textChanged.connect(self.on_text_changed)

    def on_text_changed(self, text):
        self.app.search_filter = text
        self.clear_search.setEnabled(bool(text))



import hashlib
from PySide6 import QtWidgets, QtGui

BUF_SIZE = 65536


def calculate_sha1(file_path):
    sha1 = hashlib.sha1()
    with open(file_path, "rb") as f:
        while True:
            data = f.read(BUF_SIZE)
            if not data:
                break
            sha1.update(data)
    return sha1.hexdigest()


def calculate_md5(file_path):
    md5 = hashlib.md5()
    with open(file_path, "rb") as f:
        while True:
            data = f.read(BUF_SIZE)
            if not data:
                break
            md5.update(data)
    return md5.hexdigest()


def horizontal_layout(*widgets, margins=[0, 0, 0, 0]):
    layout = QtWidgets.QHBoxLayout()
    layout.setContentsMargins(*margins)
    for widget in widgets:
        layout.addWidget(widget)
    return layout


def tags_from_text(text):
    return [tag.strip() for tag in text.split(",") if tag.strip()]


def pxmap_from_svg(svg_filepath, color="black"):
    img = QtGui.QPixmap(svg_filepath)
    qp = QtGui.QPainter(img)
    qp.setCompositionMode(QtGui.QPainter.CompositionMode_SourceIn)
    qp.fillRect(img.rect(), QtGui.QColor(color))
    qp.end()
    return img



from PySide6 import QtWidgets, QtCore
from meni.ui.flowlayout import FlowLayout


class TagRow(QtWidgets.QWidget):
    def __init__(self):
        super().__init__()
        self.app = QtCore.QCoreApplication.instance()

        self._tags = None

        self.layout = FlowLayout()
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.layout
        self.setLayout(self.layout)

    @property
    def tags(self):
        return self._tags

    @tags.setter
    def tags(self, tags):
        self._tags = tags
        self.build_tags()

    def build_tags(self):
        while self.layout.count():
            child = self.layout.takeAt(0)
            if child.widget():
                child.widget().deleteLater()

        for tag in self._tags:
            label = QtWidgets.QLabel(tag)
            label.setSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
            label.setStyleSheet(
                f"""
                background-color: {self.app.theme.tag_background};
                color: {self.app.theme.tag_foreground};

                opacity: 0.8;
                font-size: 10px;
                font-weight: bold;   

                padding: 2px;
                margin: 2px;
                border-radius: 2px;
                """
            )
            self.layout.addWidget(label)



import os
import shutil
import jsonpickle
from meni.utils import calculate_sha1
from PySide6 import QtCore
import vtkplotlib as vpl
from stl.mesh import Mesh


class Metadata(QtCore.QObject):
    changed = QtCore.Signal()

    def __init__(self):
        super().__init__()
        self.app = QtCore.QCoreApplication.instance()

        self.collections = []
        self.files = []
        self._tags = None
        self._authors = None

        self.changed.connect(self.clear_caches)
        self.changed.connect(self.validate)

    def clear_caches(self):
        self._tags = None
        self._authors = None

    def reload(self):
        self.clear_caches()
        self.changed.emit()

    def validate(self):
        # Collections without any file and with no data
        for collection in self.collections:
            if (
                collection.files == []
                and collection.attachments == []
                and collection.notes is None
                and collection.author is None
                and collection.url is None
            ):
                self.remove_collection(collection)

    @property
    def tags(self):
        if not self._tags:
            self._tags = {}
            for file in self.files:
                for tag in file.tags:
                    if tag not in self._tags:
                        self._tags[tag] = Tag(tag, 1)
                    else:
                        self._tags[tag].quantity += 1

        return list(self._tags.values())

    @property
    def authors(self):
        if not self._authors:
            self._authors = {}
            for collection in self.collections:
                author = collection.author
                if author is None:
                    author = "<No author>"
                if author not in self._authors:
                    self._authors[author] = len(collection.files)
                else:
                    self._authors[author] += len(collection.files)
        return list(self._authors.keys())

    def add_file(self, path, collection=None, name=None, tags=[]):
        new_path = shutil.copy(path, self.app.current_library)

        if name is None:
            name = os.path.basename(new_path)

        self.add_collection_if_not_exist(collection)

        hash = calculate_sha1(new_path)

        file = Local3DFile(new_path, collection, name, tags, hash)
        file.generate_thumbnail()
        self.files.append(file)
        self.changed.emit()

        return file

    def update_file(self, file, name=None, tags=None, collection=None, path=None):
        if name is not None:
            file.name = name
        if tags is not None:
            file.tags = tags
        if collection is not None:
            self.add_collection_if_not_exist(collection)
            file.collection = collection
        if path is not None:
            file.path = path

        self.changed.emit()
        return file

    def remove_file(self, file):
        try:
            os.remove(file.path)
        except FileNotFoundError:
            print(f"File not found: {file.path}")
            pass
        self.files.remove(file)
        self.changed.emit()

    def replace_file(self, file, path):
        os.remove(file.path)
        new_path = shutil.copy(path, self.app.current_library)
        file.path = new_path
        file.hash = calculate_sha1(new_path)
        file.generate_thumbnail()
        self.changed.emit()

        return file

    def add_collection(self, name, author=None, url=None, notes=None, attachments=[]):
        collection = Collection(name, author, url, notes, attachments)
        self.collections.append(collection)
        self.changed.emit()
        return collection

    def add_collection_if_not_exist(self, name, author=None, url=None, notes=None, attachments=[]):
        if self.collection_by_name(name) is None and name is not None and name.strip() != "":
            return self.add_collection(name, author, url, notes, attachments)

    def update_collection(self, collection, name=None, author=None, url=None, notes=None):
        if name is not None:
            for file in collection.files:
                file.collection = name
            collection.name = name
        if author is not None:
            collection.author = author
        if url is not None:
            collection.url = url
        if notes is not None:
            collection.notes = notes
        self.changed.emit()
        return collection

    def remove_collection(self, collection):
        for file in collection.files:
            self.remove_file(file)
        self.collections.remove(collection)
        self.changed.emit()

    def add_attachment(self, collection, path):
        new_path = shutil.copy(path, os.path.join(collection.attachments_folder, os.path.basename(path)))
        collection.attachments.append(new_path)
        self.changed.emit()
        return new_path

    def rename_attachment(self, collection, attachment, new_name):
        new_path = os.path.join(collection.attachments_folder, new_name)
        os.rename(attachment, new_path)
        collection.attachments.remove(attachment)
        collection.attachments.append(new_path)
        self.changed.emit()
        return new_path

    def remove_attachment(self, collection, attachment):
        os.remove(attachment)
        collection.attachments.remove(attachment)
        self.changed.emit()

    def commit_stage(self, stage):
        for file in stage.files:
            self.add_file(path=file.path, name=file.name, collection=stage.collection, tags=stage.tags)

    # GETTERS

    # def tag_by_name(self, name):
    #     return self._tags[name]

    # def file_by_hash(self, hash):
    #     for file in self.files:
    #         if file.hash == hash:
    #             return file

    # def file_by_name(self, name):
    #     for file in self.files:
    #         if file.name == name:
    #             return file

    # def file_by_path(self, path):
    #     for file in self.files:
    #         if file.path == path:
    #             return file

    # def file_by_library_path(self, library_path):
    #     for file in self.files:
    #         if file.library_path == library_path:
    #             return file

    def collection_by_name(self, name):
        for collection in self.collections:
            if collection.name == name:
                return collection


class JsonPickledMetadata(Metadata):
    def __init__(self):
        super().__init__()
        if QtCore.QCoreApplication.instance().current_library:
            os.makedirs(QtCore.QCoreApplication.instance().current_library, exist_ok=True)
            os.makedirs(os.path.join(QtCore.QCoreApplication.instance().current_library, "attachments"), exist_ok=True)
            os.makedirs(os.path.join(QtCore.QCoreApplication.instance().current_library, "thumbnails"), exist_ok=True)

            self.json_file = os.path.join(QtCore.QCoreApplication.instance().current_library, "metadata.json")

            self._load()
        self.changed.connect(self._save)

    def reload(self):
        self.json_file = os.path.join(QtCore.QCoreApplication.instance().current_library, "metadata.json")
        self._load()
        super().reload()

    def _save(self):
        with open(self.json_file, "w") as outfile:
            outfile.write(jsonpickle.encode({"collections": self.collections, "files": self.files}))

    def _load(self):
        self.files.clear()
        self.collections.clear()

        if os.path.exists(self.json_file):
            with open(self.json_file, "r") as infile:
                obj = jsonpickle.decode(infile.read())
                self.files.extend(obj["files"])
                self.collections.extend(obj["collections"])


class Collection:
    def __init__(self, name, notes=None, author=None, url=None, attachments=[]):
        self.name = name
        self.notes = notes
        self.author = author
        self.url = url
        self.attachments = attachments

    @property
    def files(self):
        app = QtCore.QCoreApplication.instance()
        return [file for file in app.metadata.files if file.collection_obj == self]

    @property
    def attachments_folder(self):
        return os.path.join(QtCore.QCoreApplication.instance().current_library, "attachments", self.name)

    def __str__(self):
        return f"{self.name}"

    def __repr__(self):
        return f"Collection: {self.name}"


class Local3DFile:
    def __init__(self, path, collection, name=None, tags=None, hash=None):
        self.name = name or Local3DFile._name_from_path(path)
        self.path = path
        self.collection = collection
        self.tags = tags or []
        self.hash = hash or calculate_sha1(path)

    def _name_from_path(path):
        name = os.path.basename(path)
        name = os.path.splitext(name)[0]
        name = name.replace("_", " ")
        name = name.replace("-", " ")
        name = " ".join([word.capitalize() for word in name.split()])
        return name

    @property
    def absolute_path(self):
        return os.path.abspath(self.path)

    @property
    def thumbnail_file(self):
        return os.path.join(QtCore.QCoreApplication.instance().current_library, "thumbnails", f"{self.hash}.png")

    @property
    def collection_obj(self):
        metadata = QtCore.QCoreApplication.instance().metadata
        collection = metadata.collection_by_name(self.collection)
        if collection is None and self.collection is not None and self.collection.strip() != "":
            collection = metadata.add_collection(self.collection)
        return collection

    def generate_thumbnail(self):
        mesh = Mesh.from_file(self.path)
        fig = vpl.figure()
        fig.background_opacity = 0
        vpl.mesh_plot(mesh, color="#8ec07c", fig=fig)
        # vpl.view(focal_point=[0, 0, 0], camera_position=[-50, -50, 50], fig=fig)
        vpl.reset_camera(fig=fig)
        vpl.save_fig(self.thumbnail_file, off_screen=True, fig=fig)
        vpl.close(fig=fig)

    def __str__(self):
        return f"{self.name} ({self.path})"

    def __repr__(self):
        return f"Local3DFile: {self.path}"


class Tag:
    def __init__(self, name, quantity):
        self.name = name
        self.quantity = quantity

    def __str__(self):
        return f"{self.name} ({self.quantity})"

    def __repr__(self):
        return f"Tag: {self.name}"


class Stage:
    def __init__(self):
        self.files = []
        self.collection = None
        self.tags = []

    def __str__(self):
        return f"{self.collection} ({len(self.files)} files)"

    def __repr__(self):
        return f"Stage: {self.collection}"



from PySide6 import QtWidgets, QtCore
from meni.ui.tagrow import TagRow
from meni.ui.common import DockTitleBar, IconLabel
from stl.mesh import Mesh
import vtkplotlib as vpl
import qtawesome as qta


class ViewerDock(QtWidgets.QDockWidget):
    def __init__(self, parent):
        super().__init__("Viewer", objectName="viewer", parent=parent)

        self.app = QtCore.QCoreApplication.instance()
        self.mesh = None

        self.setTitleBarWidget(DockTitleBar("Viewer", clicked=self.close))

        self.setFeatures(QtWidgets.QDockWidget.DockWidgetFeature.DockWidgetMovable | QtWidgets.QDockWidget.DockWidgetFeature.DockWidgetClosable)
        self.setAllowedAreas(QtCore.Qt.DockWidgetArea.AllDockWidgetAreas)

        self.layout = QtWidgets.QVBoxLayout()
        self.layout.setSpacing(0)
        self.layout.setContentsMargins(5, 5, 5, 5)

        # Title
        self.title = QtWidgets.QLabel()
        self.title.setStyleSheet("font-size: 20px; font-weight: bold;text-decoration: underline;")
        self.title.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        self.layout.addWidget(self.title)

        self.layout.addSpacing(10)

        # Path
        self.path = IconLabel(qta_id="fa5s.file", icon_size=12)
        self.path.setStyleSheet("font-size: 10px; opacity: 0.8")
        self.layout.addWidget(self.path)

        # Collection
        self.collection = IconLabel(qta_id="fa5s.layer-group", icon_size=12, final_stretch=True)
        self.collection.setStyleSheet("font-size: 10px; opacity: 0.8")
        self.collection.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        self.layout.addWidget(self.collection)

        # Tags
        self.tagrow = TagRow()
        self.tagrow.setContentsMargins(0, 10, 0, 0)
        self.layout.addWidget(self.tagrow)

        # Mesh viewer
        self.fig = vpl.QtFigure2()
        self.fig.vl.setContentsMargins(0, 0, 0, 0)
        self.fig.background_color = self.app.theme.main_background
        self.show_stl(None)
        self.layout.addWidget(self.fig)

        # Set layout to widget to dock
        widget = QtWidgets.QWidget()
        widget.setLayout(self.layout)
        self.setWidget(widget)

        # Connect signals
        self.app.metadata.changed.connect(lambda: self.on_selected_file_changed([], self.app.last_selected_file))
        self.app.selected_files_changed.connect(self.on_selected_file_changed)

        # Starting values
        self.on_selected_file_changed([], None)

    def on_selected_file_changed(self, files, last):
        self.path.setVisible(last is not None)
        self.collection.setVisible(last is not None and last.collection_obj is not None)

        if last:
            self.title.setText(last.name)
            self.path.setText(last.path)
            self.collection.setText(last.collection_obj.name if last.collection_obj else "")
            self.tagrow.tags = last.tags
            self.show_stl(last.path)
        else:
            self.title.setText("")
            self.path.setText("")
            self.collection.setText("")
            self.tagrow.tags = []
            self.show_stl(None)

    def show_stl(self, stl):
        if self.mesh:
            self.fig.remove_plot(self.mesh)

        if stl is None:
            self.fig.update()
            return

        self.mesh = vpl.mesh_plot(Mesh.from_file(stl), color=self.app.theme.model_color, fig=self.fig)
        vpl.reset_camera(fig=self.fig)
        self.fig.update()
